/* x10bot -- main source file.
 * Copyright 2001-2002 Daniel P. Stasinski <mooooooo@scriptkitties.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <x10_api.h>
#include <tcl.h>
#include <tclDecls.h>
#include <getopt.h>
#include "x10bot.h"

int running = 0;
int sockets[MAXSOCKETS];
int accept_socket = 0, sock = 0;
struct x10_status_t *x10_status = NULL;
x10module modules[MAXMODULES];
event events[MAXEVENTS];
Tcl_Interp *interp;
int port = PORT;
int foreground = 0;
char *logfile = NULL;
char *device = NULL;
time_t uptime;
int eventhandle = 0;
double latitude = 0.0;
double longitude = 0.0;
COMMANDLIST *addedcommands = NULL;
int thissocket = 0;

char *status_to_string[] = {
    "UNKNOWN",
    "ON",
    "OFF",
    "DIM_BRIGHT"
};

char *cmd_to_string[] = {
    "CMD_ALL_UNITS_OFF",
    "CMD_ALL_LIGHTS_ON",
    "CMD_ON",
    "CMD_OFF",
    "CMD_DIM",
    "CMD_BRIGHT",
    "CMD_ALL_LIGHTS_OFF",
    "CMD_EXTENDED_CODE",
    "CMD_HAIL_REQUEST",
    "CMD_HAIL_AKNOWLEDGE",
    "CMD_PRESET_DIM1",
    "CMD_PRESET_DIM2",
    "CMD_EXTENDED_DATA_TRANSFER",
    "CMD_STATUS_ON",
    "CMD_STATUS_OFF",
    "CMD_STATUS_REQUEST"
};

char *asciitime(const time_t * timep)
{
    int size;
    char *t;

    if ((timep == NULL) || (*timep == 0))
        return "null";

    t = ctime(timep);
    size = strlen(t);
    t[size - 1] = 0;

    return t;
}

void fatal_error(char *msg)
{
    fprintf(stderr, msg);
    exit(1);
}

void camera_group_off(int module_id, struct x10_status_t *status)
{
    int x;

    if (modules[module_id].group != NULL) {
        if (!strcasecmp(modules[module_id].group, "camera")) {
            for (x = 0; x < MAXMODULES; x++) {
                if (modules[x].group == NULL)
                    continue;
                if (strcasecmp(modules[x].group, "camera"))
                    continue;
                if (x == module_id)
                    continue;
                x10_log_event((x / 16) + 'A', (x % 16) + 1, X10_CMD_OFF, x10_status);
            }
        }
    }
}

int gethouse(char *module_id)
{
    char house;

    house = toupper(module_id[0]);

    if (house < 'A' || house > 'P')
        return (-1);

    return (house - 'A');
}

int getmodule(char *module_id)
{
    int x;
    char house;
    int unit;

    for (x = 0; x < MAXMODULES; x++) {
        if (modules[x].alias != NULL) {
            if (!strcasecmp(modules[x].alias, module_id))
                return (x);
        }
    }

    if (strlen(module_id) < 2)
        return (-1);

    house = toupper(module_id[0]);
    unit = atoi(&module_id[1]);

    if (house < 'A' || house > 'P' || unit < 1 || unit > 16)
        return (-1);

    return (((house - 'A') * 16) + (unit - 1));
}

char *cmd2string(int command)
{
    return (cmd_to_string[command]);
}

int string2cmd(char *command)
{
    int i;
    for (i = 0; i < 16 && strcasecmp(command, cmd_to_string[i]); i++);
    return (i == 16 ? -1 : i);
}

char *status2string(int status)
{
    return (status_to_string[status]);
}

int string2status(char *status)
{
    int i;
    for (i = 0; i < 4 && strcasecmp(status, status_to_string[i]); i++);
    return (i == 4 ? -1 : i);
}


void sighup(int signum)
{
    x10_tcl_init();
}

void sigquit(int signum)
{
    running = 0;
    return;
}

void error_handler(int code, char *arg, struct x10_status_t *status)
{
    char reply[BUFLEN];
    sprintf(reply, "ERROR %d\n", code);
    send_to_all(reply);
}

void function_hook(unsigned char house, int unit, int function, int dim_level, struct x10_status_t *status)
{

    char reply[BUFLEN];
    int module_id;
    char *info, *errorInfo;

    if (house < 'A' || house > 'P' || unit < 1 || unit > 16) {
        sprintf(reply, "ERROR %c%d %d %d\n", house, unit, function, dim_level);
        send_to_all(reply);
        return;
    }

    x10_log_event(house, unit, function, status);

    switch (function) {
    case 0:
        module_id = 256;
        break;
    case 1:
    case 6:
        module_id = 257;
        break;
    default:
        module_id = ((house - 'A') * 16) + (unit - 1);
    }

    if (modules[module_id].alias != NULL)
        sprintf(reply, "EVENT %s %s %d\n", modules[module_id].alias, cmd2string(function), dim_level);
    else
        sprintf(reply, "EVENT %c%d %s %d\n", house, unit, cmd2string(function), dim_level);

    if (function == X10_CMD_ON)
        camera_group_off(module_id, status);

    send_to_all(reply);

    if (modules[module_id].process != NULL) {
        if (modules[module_id].alias != NULL)
            sprintf(reply, "%s %s %s %d\n", modules[module_id].process, modules[module_id].alias, cmd2string(function), dim_level);
        else
            sprintf(reply, "%s %c%d %s %d\n", modules[module_id].process, house, unit, cmd2string(function), dim_level);

        if (Tcl_Eval(interp, reply) == TCL_ERROR) {
            sprintf(reply, "ERROR (%s) %s\n", modules[module_id].process, interp->result);
            send_to_all(reply);
            errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
            if (errorInfo != NULL) {
                info = strtok(errorInfo, "\n");
                while (info != NULL) {
                    sprintf(reply, "DEBUG   %s\n", info);
                    send_to_all(reply);
                    info = strtok(NULL, "\n");
                }
            }
        }
    }
}

int insert_event(int seconds, int mode, int once, int thread, int quiet, int repeat, char *script)
{
    int x, size;
    struct tm *tp;
    time_t now;
    time_t timestamp;
    int interval;

    if (script == NULL)
        return (-1);

    now = time(NULL);

    tp = localtime(&now);

    switch (mode) {
    case SCHED_PROC:
        interval = seconds;
        timestamp = now + seconds;
        break;
    case SCHED_DAILY:
    case SCHED_WEEKDAY:
    case SCHED_WEEKEND:
        interval = 86400;
        tp->tm_sec = 0;
        tp->tm_hour = (seconds / 3600);
        tp->tm_min = (seconds % 3600) / 60;
        timestamp = mktime(tp);
        break;
    case SCHED_WEEKLY:
        interval = (86400 * 7);
        tp->tm_sec = 0;
        tp->tm_wday = (seconds / 86400);
        tp->tm_hour = (seconds / 3600) % 24;
        tp->tm_min = (seconds % 3600) / 60;
        timestamp = mktime(tp);
        break;
    case SCHED_ONCE:
        interval = 0;
        timestamp = seconds;
        break;
    }

    if (timestamp <= now)
        timestamp += interval;

    for (x = 0; x < MAXEVENTS; x++) {
        if (events[x].mode == SCHED_NEVER) {
            eventhandle = (eventhandle % INT_MAX) + 1;  /* eventhandle will never be 0 */
            events[x].handle = eventhandle;
            events[x].timestamp = timestamp;
            events[x].interval = interval;
            events[x].mode = mode;
            events[x].once = once;
            events[x].thread = thread;
            events[x].quiet = quiet;
            events[x].repeat = repeat;
            size = strlen(script);
            if (size) {
                events[x].script = Tcl_Alloc(size + 1);
                strcpy(events[x].script, script);
            } else
                events[x].script = NULL;
            return (events[x].handle);
        }
    }
    return (-1);
}

int do_event(int lastevent)
{

    char reply[BUFLEN];
    time_t now;
    struct tm *timeptr;
    int x;
    char *info, *errorInfo;

    now = time(NULL);
    timeptr = localtime(&now);

    for (x = 0; x < MAXEVENTS; x++) {
        lastevent = (lastevent + 1) % MAXEVENTS;
        if (events[lastevent].mode == SCHED_NEVER)
            continue;
        if (now >= events[lastevent].timestamp) {

            if (events[lastevent].repeat > 0)
                events[lastevent].repeat--;

            if (events[lastevent].mode == SCHED_WEEKDAY) {
                if ((timeptr->tm_wday == 0) || (timeptr->tm_wday == 6)) {
                    events[lastevent].timestamp += 86400;
                    break;
                }
            } else if (events[lastevent].mode == SCHED_WEEKEND) {
                if ((timeptr->tm_wday > 0) && (timeptr->tm_wday < 6)) {
                    events[lastevent].timestamp += 86400;
                    break;
                }
            }
            if (!events[lastevent].quiet) {
                sprintf(reply, "SCHEDULED %s\n", events[lastevent].script);
                send_to_all(reply);
            }
            if (Tcl_Eval(interp, events[lastevent].script) == TCL_ERROR) {
                sprintf(reply, "ERROR (%s) %s\n", events[lastevent].script, interp->result);
                send_to_all(reply);
                errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
                if (errorInfo != NULL) {
                    info = strtok(errorInfo, "\n");
                    while (info != NULL) {
                        sprintf(reply, "DEBUG   %s\n", info);
                        send_to_all(reply);
                        info = strtok(NULL, "\n");
                    }
                }
            }
            if (events[lastevent].once || events[lastevent].repeat == 0) {
                if (events[lastevent].script != NULL) {
                    Tcl_Free(events[lastevent].script);
                }
                events[lastevent].mode = SCHED_NEVER;
                events[lastevent].script = NULL;
            } else {
                events[lastevent].timestamp += events[lastevent].interval;
            }
            break;
        }
    }

    return (lastevent);

}

void usage(void)
{
    printf("  : options:\n");
    printf("    -h            this help message\n");
    printf("    -f            foreground mode (defaults to daemon mode)\n");
    printf("    -d <device>   device to talk to CM11A on (default: /dev/ttyS0)\n");
    printf("    -l <logfile>  file name to log commands and events to\n");
    printf("    -p <port>     port to listen on (default: 4466)\n");
}

int main(int argc, char *argv[])
{
    char c;
    int x, f, lastevent;

    printf("x10bot\n");

    uptime = time(NULL);

    /* parse the command arguments */
    opterr = 0;
    while ((c = getopt(argc, argv, "hfd:p:l:")) != -1) {
        switch (c) {
        case 'h':
            usage();
            goto done;
        case 'f':
            foreground = 1;
            break;
        case 'd':
            if (!optarg) {
                usage();
                goto done;
            }
            device = Tcl_Alloc(strlen(optarg) + 1);
            strcpy(device, optarg);
            break;
        case 'l':
            if (!optarg) {
                usage();
                goto done;
            }
            logfile = strdup(optarg);
            break;
        case 'p':
            if (!optarg) {
                usage();
                goto done;
            }
            port = atoi(optarg);
            break;
        case '?':
            fprintf(stderr, "ERR: Unknown option `-%c'.\n", optopt);
            usage();
            goto done;
        default:
            usage();
            goto done;
        }
    }

    if (!device) {
        device = Tcl_Alloc(strlen(SERIAL) + 1);
        strcpy(device, SERIAL);
    }

    for (x = 0; x < MAXMODULES; x++) {
        modules[x].alias = NULL;
        modules[x].description = NULL;
        modules[x].process = NULL;
        modules[x].group = NULL;
    }

    for (x = 0; x < MAXEVENTS; x++) {
        events[x].script = NULL;
        events[x].mode = SCHED_NEVER;
    }

    for (x = 0; x < MAXSOCKETS; x++)
        sockets[x] = -1;

    signal(SIGHUP, sighup);
    signal(SIGTERM, sigquit);

    if (!(x10_status = x10_open(device, error_handler, function_hook, TIMEOUT_READ_CM11A))) {
        printf("x10 Init Error.\n");
        return (0);
    }

    x10_set_status_interface(x10_status);

    interp = Tcl_CreateInterp();
    Tcl_Init(interp);

    x10_tcl_init();

    send_to_all("STARTUP\n");

    accept_socket = init_server();

    if (!foreground) {
        switch (f = fork()) {
        case -1:
            printf("fatal error: unable to fork()");
            exit(-1);
        case 0:
            setsid();
            break;
        default:
            exit(0);
        }
    }

    lastevent = 0;
    running = 1;

    while (running) {
        x10_listen_to_events(x10_status, PERIOD_POLL_CM11A);
        lastevent = do_event(lastevent);
        Tcl_DoOneEvent(TCL_DONT_WAIT);
        cycle_on_connections(accept_socket);
    }

    send_to_all("SHUTDOWN\n");

    for (x = 0; x < MAXSOCKETS; x++) {
        if (sockets[x] != -1) {
            shutdown(sockets[x], 2);
            close(sockets[x]);
        }
    }

    shutdown(accept_socket, 2);

    close(accept_socket);

    x10_close(x10_status);

    Tcl_DeleteInterp(interp);

    for (x = 0; x < MAXMODULES; x++) {
        if (modules[x].alias != NULL)
            Tcl_Free(modules[x].alias);
        if (modules[x].description != NULL)
            Tcl_Free(modules[x].description);
        if (modules[x].process != NULL)
            Tcl_Free(modules[x].process);
        if (modules[x].group != NULL)
            Tcl_Free(modules[x].group);
    }

    for (x = 0; x < MAXEVENTS; x++) {
        events[x].mode = SCHED_NEVER;
        if (events[x].script != NULL)
            Tcl_Free(events[x].script);
    }

  done:

    if (device)
        Tcl_Free(device);

    if (logfile)
        Tcl_Free(logfile);

    sleep(1);

    exit(1);
}
